//------------------------------------------------------------------------------
// File: RateChanger.cpp
//------------------------------------------------------------------------------

#include <streams.h>
#include "dxutil.h"

#include "RateChanger.h"

/*
* July 5, 1991
* Copyright 1991 Lance Norskog And Sundry Contributors
* This source code is freely redistributable and may be used for
* any purpose.  This copyright notice must be maintained. 
* Lance Norskog And Sundry Contributors are not responsible for 
* the consequences of using this software.
*/

/*
* Modified for use in HawkVoiceDI by Phil Frisbie, Jr.
* Sound Tools rate change effect file.
* http://www.hawksoft.com/hawkvoice/
*/

/*
* Strongly modified for ReClock by OGO :)
*
*/

/*
* Least Common Multiple Linear Interpolation 
*
* Find least common multiple of the two sample rates.
* Construct the signal at the LCM by interpolating successive
* input samples as straight lines.  Pull output samples from
* this line at output rate.
*
* Of course, actually calculate only the output samples.
*
* LCM must be 32 bits or less.  Two prime number sample rates
* between 32768 and 65535 will yield a 32-bit LCM, so this is 
* stretching it.
*/

/*
* Algorithm:
*	
*	Generate a master sample clock from the LCM of the two rates.
*	Interpolate linearly along it.	Count up input and output skips.
*
*	Input:	 |inskip |		 |		 |		 |		 |
*																		
*																		
*																		
*	LCM:	 |	 |	 |	 |	 |	 |	 |	 |	 |	 |	 |
*																		
*																		
*																		
*	Output:  |	outskip  |			 |			 | 
*
*																		
*/

/* here for linear interp.	might be useful for other things */

static ULONG_UNIT gcd(ULONG_UNIT a, ULONG_UNIT b) 
{
	if (b == 0)
		return a;
	else
		return gcd(b, a % b);
}

static ULONG_UNIT lcm(ULONG_UNIT a, ULONG_UNIT b) 
{
	return (a * b) / gcd(a, b);
}

RateChanger::RateChanger(int channelCount, int inrate, int outrate)
{ 
	m_iRate = m_oRate = 1;
	m_inskip = m_outskip = 0; 
	m_channels = 0;
	SetChannelCount(channelCount);
	SetRate(inrate, outrate);
}

RateChanger::~RateChanger()
{
	SAFE_DELETE_ARRAY(m_channels);
}

void RateChanger::SetChannelCount(int channelCount)
{
	SAFE_DELETE_ARRAY(m_channels);
	m_channelCount = channelCount;
	m_channels = new Channel[m_channelCount];
}

/*
* Prepare processing.
*/
void RateChanger::SetRate(int inrate, int outrate)
{
	m_lcmrate = lcm((ULONG_UNIT) inrate, (ULONG_UNIT) outrate);
	/* Cursory check for LCM overflow.	
	* If both rate are below 65k, there should be no problem.
	* 16 bits x 16 bits = 32 bits, which we can handle.
	*/
	m_inskip = m_lcmrate / inrate;
	m_outskip = m_lcmrate / outrate; 
	m_iRate = inrate;
	m_oRate = outrate;
	for (int i=0; i<m_channelCount; i++)
	{
		m_channels[i].m_intot = m_channels[i].m_outtot = 0;
	}
}

/*
* Processed signed short samples from ibuf to obuf.
* Return number of samples processed.
*/
void RateChanger::ProcessFlow16(int channel, short *ibuf, short *obuf, int inlen, int *outlen, int adv)
{
	short *istart = ibuf;
	Channel* l_channel = &m_channels[channel];
	
	*outlen = 0;
	
	if (l_channel->m_total == 0)
	{
		/* Emit first sample.  We know the fence posts meet. */
		*obuf = *ibuf;
		/* Count up until have right input samples */
		l_channel->m_lastsamp = *obuf;
		obuf += adv;
		(*outlen)++;
		l_channel->m_total = 1;
		/* advance to second output */
		l_channel->m_outtot += m_outskip;
	}
	
	while ((l_channel->m_intot + m_inskip) <= l_channel->m_outtot)
	{
		l_channel->m_lastsamp = *ibuf;
		ibuf += adv;
		l_channel->m_intot += m_inskip;
		if ((ibuf - istart)/adv == inlen)
			return;
	}
	/* long samples with high LCM's overrun counters! */
	if (l_channel->m_outtot == l_channel->m_intot)
		l_channel->m_outtot = l_channel->m_intot = 0;
	
	for(;;) 
	{
		LONG_UNIT osl = l_channel->m_lastsamp + (((LONG_UNIT)*ibuf - l_channel->m_lastsamp) * (LONG_UNIT)(l_channel->m_outtot - l_channel->m_intot)) / m_inskip;
		*obuf = (short)osl;
		obuf += adv;
		(*outlen)++;
		/* advance to next output */
		l_channel->m_outtot += m_outskip;
		/* advance input range to span next output */
		while ((l_channel->m_intot + m_inskip) <= l_channel->m_outtot)
		{
			l_channel->m_lastsamp = *ibuf;
			ibuf += adv;
			l_channel->m_intot += m_inskip;
			if ((ibuf - istart)/adv == inlen)
				return;
		}
		/* long samples with high LCM's overrun counters! */
		if (l_channel->m_outtot == l_channel->m_intot)
			l_channel->m_outtot = l_channel->m_intot = 0;
	}
}

/*
* Processed signed signed char samples from ibuf to obuf.
* Return number of samples processed.
*/
void RateChanger::ProcessFlow8(int channel, unsigned char *ibuf, unsigned char *obuf, int inlen, int *outlen, int adv)
{
	unsigned char *istart = ibuf;
	Channel* l_channel = &m_channels[channel];
	
	*outlen = 0;
	
	if (l_channel->m_total == 0)
	{
		/* Emit first sample.  We know the fence posts meet. */
		*obuf = *ibuf;
		/* Count up until have right input samples */
		l_channel->m_lastsamp = *obuf;
		obuf += adv;
		(*outlen)++;
		l_channel->m_total = 1;
		/* advance to second output */
		l_channel->m_outtot += m_outskip;
	}
	
	while ((l_channel->m_intot + m_inskip) <= l_channel->m_outtot)
	{
		l_channel->m_lastsamp = *ibuf;
		ibuf += adv;
		l_channel->m_intot += m_inskip;
		if ((ibuf - istart)/adv == inlen)
			return;
	}
	/* long samples with high LCM's overrun counters! */
	if (l_channel->m_outtot == l_channel->m_intot)
		l_channel->m_outtot = l_channel->m_intot = 0;
	
	for(;;) 
	{
		LONG_UNIT osl = l_channel->m_lastsamp + (((LONG_UNIT)*ibuf - l_channel->m_lastsamp) * (LONG_UNIT)(l_channel->m_outtot - l_channel->m_intot)) / m_inskip;
		*obuf = (unsigned char)osl;
		obuf += adv;
		(*outlen)++;
		/* advance to next output */
		l_channel->m_outtot += m_outskip;
		/* advance input range to span next output */
		while ((l_channel->m_intot + m_inskip) <= l_channel->m_outtot)
		{
			l_channel->m_lastsamp = *ibuf;
			ibuf += adv;
			l_channel->m_intot += m_inskip;
			if ((ibuf - istart)/adv == inlen)
				return;
		}
		/* long samples with high LCM's overrun counters! */
		if (l_channel->m_outtot == l_channel->m_intot)
			l_channel->m_outtot = l_channel->m_intot = 0;
	}
}






